'''
Buckets hold groups of tqdm instances.
    They are used to aggregate like metrics over time.
    For example if there are two concurrent 'download' tqdms and a single
      'upload' tqdm these would be grouped into two buckets.
'''
from datetime import datetime

from tqdmpromproxy.snapshot import TqdmSnapshot


class PrometheusBucket:
    '''
    Collection of all grouped metrics'''

    def __init__(self, bucket_key: str, item_scale: str, retired_attrs: list[str], current_attrs: list[str]):
        self.bucket_key = bucket_key
        self.item_scale = item_scale
        self.current_attrs = list.copy(current_attrs)
        self.retired_attrs = list.copy(retired_attrs)

        self.aggregated = dict([(attr, 0) for attr in self.retired_attrs])

        self.known_instances: dict[TqdmSnapshot, datetime] = {}
        self.retired_instances = 0

    def matches(self, instance: TqdmSnapshot):
        return instance.bucket_key() == self.bucket_key

    def upsert(self, instance: TqdmSnapshot):
        if instance.identity() not in [i.identity() for i in self.known_instances]:
            self.known_instances[instance] = 0

        self.known_instances[instance] = datetime.now()

    # pylint: disable=unnecessary-dunder-call

    def to_prometheus_lines(self):
        '''Return the metrics as a prometheus string'''
        yield f"# TQDM group {self.bucket_key} "
        yield f"{self.bucket_key}_active_count {len(self.known_instances)}"
        yield f"{self.bucket_key}_finished_count {self.retired_instances}"

        yield f"## Individual properties with scale {self.item_scale}"
        for prop in self.current_attrs:

            val = sum([instance.__getattribute__(prop) or 0
                      for instance in self.known_instances])

            if prop in self.retired_attrs:
                val += self.aggregated[prop]

            yield f"{self.bucket_key}_{prop}_{self.item_scale} {val}"

    # pylint: disable=unnecessary-dunder-call
    def retire(self, instance: TqdmSnapshot):
        '''Move an instance from the active to the retired list'''
        if instance.identity() in [i.identity() for i in self.known_instances]:
            for prop in self.retired_attrs:
                self.aggregated[prop] += instance.__getattribute__(prop)

            del self.known_instances[instance]

            self.retired_instances += 1

    def prune(self, max_age: int):
        '''Remove any instances that are no longer active'''
        to_remove = []
        for key, value in self.known_instances.items():
            if (datetime.now() - value).total_seconds() > max_age:
                to_remove.append(key)

        for key in to_remove:
            self.retire(key)

        return len(to_remove)

    @classmethod
    def from_instance(cls, snapshot: TqdmSnapshot, retired=None, current=None):

        default_retired = ['completed']
        default_current = ['completed', 'total', 'rate']

        if retired is None:
            retired = default_retired

        if current is None:
            current = default_current

        snap = cls(snapshot.bucket_key(), snapshot.unit, retired, current)
        snap.upsert(snapshot)

        return snap
